#!/bin/ksh

debug=0

awk '

BEGIN {
    next_slot_loc = 0;
    debug = '$debug';
}

function slot_alloc(ref_type) {
    new_slot = -1;

    for (new_slot in slot_free) {
        break;
    }

    if (new_slot == -1) {
        new_slot = next_slot_loc;
        ++ next_slot_loc;
    } else {
        delete slot_free[new_slot];
    }

    if (ref_type ~ /temp/) {
        slot_temp[new_slot] = 1;
        slot_status[new_slot] = "temp";
    }

    if (ref_type ~ /var/) {
        slot_var[new_slot] = 1;
        slot_status[new_slot] = "var";
    }

    return new_slot;
}

function slot_map() {
    printf("INSTR:print \"next_slot_loc=\",%s,\"\\n\"\n", next_slot_loc);
    for (i=0; i<next_slot_loc; ++i) {
        if (slot_status[i] == "var") {
            printf("INSTR:print \"slot[%s]=%s:%s\\n\"\n",
                   i, slot_status[i], slot_var_name[i]);
        } else {
            printf("INSTR:print \"slot[%s]=%s\\n\"\n", i, slot_status[i]);
        }
    }
}

function slot_gc() {
    for (slot_loc in slot_temp) {
        slot_status[slot_loc] = "free";
        slot_free[slot_loc] = 1;
    }

    delete slot_temp;
}

# QUIT:

$1 ~ /^quit$/ {
    printf("INSTR:quit\n");
    print "DONE:";
    fflush();
    next;
}

# SLOT_MAP:

$1 ~ /^slot$/ {
    slot_map();
    printf("INSTR:ignore = prompt()\n");
    print "DONE:";
    fflush();
    next;
}

# PROMPT:

$1 ~ /^$/ {
    slot_gc();
    printf("INSTR:ignore = prompt()\n");
    print "DONE:";
    fflush();
    next;
}

# DISPLAY:

$1 ~ /^@[0-9]+$/ {
    i = $1;
    sub(/^@/,"",i);
    printf("INSTR:ignore = display(%s)\n",i);
    print "EXPR:";
    print "DONE:";
    fflush();
    next;
}

# PARENTHESIS:

$1 ~ /\(@[0-9]+\)/ {
    if (debug) printf("INSTR:parenthesis\n");

    expr = $1;
    expr_len = length(expr);

    replace_pos = match(expr, /\(@[0-9]+\)/);
    replace_len = RLENGTH;

    begin = 1;
    len = replace_pos - 1;
    head = substr(expr, begin, len);

    begin += len + 1;
    len = replace_len - 2;
    subexpr = substr(expr, begin, len);

    begin += replace_len - 1;
    len = expr_len - begin + 1;
    tail = substr(expr, begin, len);

    if (debug) printf("INSTR:head = <%s>\n", head);
    if (debug) printf("INSTR:subexpr = <%s>\n", subexpr);
    if (debug) printf("INSTR:tail = <%s>\n", tail);

    new_expr = head subexpr tail;
    print "EXPR:" new_expr;
    print "DONE:";
    fflush();
    next;
}

# VARIABLE:

$1 ~ /[a-z]+/ {
    if (debug) printf("INSTR:variable\n");

    expr = $1;
    expr_len = length(expr);

    replace_pos = match(expr, /[a-z]+/);
    replace_len = RLENGTH;

    begin = 1;
    len = replace_pos - 1;
    head = substr(expr, begin, len);

    begin += len;
    len = replace_len;
    var_name = substr(expr, begin, len);

    begin += replace_len;
    len = expr_len - begin + 1;
    tail = substr(expr, begin, len);

    if (debug) printf("INSTR:head = <%s>\n", head);
    if (debug) printf("INSTR:variable = <%s>\n", var_name);
    if (debug) printf("INSTR:tail = <%s>\n", tail);

    if (var_name in var_slot) {
        reference = "@" var_slot[var_name];
    } else {
        slot = slot_alloc("var");
        var_slot[var_name] = slot;
        slot_var_name[slot] = var_name;
        reference = "@" slot;
    }
    new_expr = head reference tail;
    print "EXPR:" new_expr;
    print "DONE:";
    fflush();
    next;
}

# INITIALIZE:

# The block below picks up literal values made up of digit strings
# which do *not* begin with a @.  This block handles digit strings
# that begin somewhere inside the expression (this is the first case)
# as well as digit strings that begin at the beginning of the expression
# (this is the second case).

$1 ~ /[^@0-9][0-9]+/ || $1 ~ /^[0-9]+/ {
    if (debug) printf("INSTR:initialize\n");

    expr = $1;
    expr_len = length(expr);

    if ($1 ~ /[^@0-9][0-9]+/) {
	literal_pos = match(expr, /[^@0-9][0-9]+/);
	literal_len = RLENGTH - 1;

	begin = 1;
	len = literal_pos;
	head = substr(expr, begin, len);

	begin += literal_pos;
	len = literal_len;
	literal = substr(expr, begin, len);

	begin += literal_len;
	len = expr_len - begin + 1;
	tail = substr(expr, begin, len);
    } else {
	literal_pos = match(expr, /^[0-9]+/);
	literal_len = RLENGTH;

	head = "";

	begin = 1;
	len = literal_len;
	literal = substr(expr, begin, len);

	begin += literal_len;
	len = expr_len - begin + 1;
	tail = substr(expr, begin, len);
    }

    if (debug == 1) printf("INSTR:head = <%s>\n", head);
    if (debug == 1) printf("INSTR:literal = <%s>\n", literal);
    if (debug == 1) printf("INSTR:tail = <%s>\n", tail);

    slot = slot_alloc("temp");
    printf("INSTR:ignore = initialize(%s,%s)\n", slot, literal);

    reference = "@" slot;
    new_expr = head reference tail;
    print "EXPR:" new_expr;
    print "DONE:";
    fflush();
    next;
}

# COPY:

$1 ~ /^@[0-9]+=@[0-9]+$/ {
    if (debug) printf("INSTR:copy\n");

    lvar = $1;
    sub(/=.*$/, "", lvar);
    sub(/@/, "", lvar);
    rvar = $1;
    sub(/^.*=/, "", rvar);
    sub(/@/, "", rvar);

    if (debug == 1) printf("INSTR:lvar = <%s>\n", lvar);
    if (debug == 1) printf("INSTR:rvar = <%s>\n", rvar);

    if (slot_status[lvar] == "var") {
        printf("INSTR:ignore = copy(%s,%s)\n", lvar, rvar);
    } else {
        printf("INSTR:print \"error: left operand is not a variable\"\n");
    }
    print "EXPR:";
    print "DONE:";
    fflush();
    next;
}

# PRODUCT:

$1 ~ /@[0-9]+\*@[0-9]+/ {
    if (debug) printf("INSTR:product\n");

    expr = $1;
    expr_len = length(expr);

    replace_pos = match(expr, /@[0-9]+\*@[0-9]+/) - 1;
    replace_len = RLENGTH;

    begin = 1;
    len = replace_pos;
    head = substr(expr, begin, len);

    begin += replace_pos;
    len = replace_len;
    replace = substr(expr, begin, len);

    begin += replace_len;
    len = expr_len - begin + 1;
    tail = substr(expr, begin, len);

    if (debug) printf("INSTR:head = <%s>\n", head);
    if (debug) printf("INSTR:replace = <%s>\n", replace);
    if (debug) printf("INSTR:tail = <%s>\n", tail);

    split(replace, part, /\*/);
    sub(/^@/, "", part[1]);
    sub(/^@/, "", part[2]);

    slot = slot_alloc("temp");
    printf("INSTR:ignore = product(%s,%s,%s)\n", slot, part[1], part[2]);

    reference = "@" slot;
    new_expr = head reference tail;
    print "EXPR:" new_expr;
    print "DONE:";
    fflush();
    next;
}

# SUM:

$1 ~ /@[0-9]+\+@[0-9]+/ {
    if (debug) printf("INSTR:sum\n");

    expr = $1;
    expr_len = length(expr);

    replace_pos = match(expr, /@[0-9]+\+@[0-9]+/) - 1;
    replace_len = RLENGTH;

    begin = 1;
    len = replace_pos;
    head = substr(expr, begin, len);

    begin += replace_pos;
    len = replace_len;
    replace = substr(expr, begin, len);

    begin += replace_len;
    len = expr_len - begin + 1;
    tail = substr(expr, begin, len);

    if (debug) printf("INSTR:head = <%s>\n", head);
    if (debug) printf("INSTR:replace = <%s>\n", replace);
    if (debug) printf("INSTR:tail = <%s>\n", tail);

    split(replace, part, /\+/);
    sub(/^@/, "", part[1]);
    sub(/^@/, "", part[2]);

    slot = slot_alloc("temp");
    printf("INSTR:ignore = sum(%s,%s,%s)\n", slot, part[1], part[2]);

    reference = "@" slot;
    new_expr = head reference tail;
    print "EXPR:" new_expr;
    print "DONE:";
    fflush();
    next;
}

# NEGATE:

$1 ~ /^-@[0-9]+/ || $1 ~ /[^0-9]-@[0-9]+/ {
    if (debug) printf("INSTR:negate\n");

    expr = $1;
    expr_len = length(expr);

    replace_pos = match(expr, /^-@[0-9]+/);
    if (replace_pos >= 1) {
        replace_pos -= 1;
        replace_len = RLENGTH;
    } else {
        replace_pos = match(expr, /[^0-9]-@[0-9]+/);
        replace_len = RLENGTH - 1;
    }

    begin = 1;
    len = replace_pos;
    head = substr(expr, begin, len);

    begin += replace_pos;
    len = replace_len;
    replace = substr(expr, begin, len);

    begin += replace_len;
    len = expr_len - begin + 1;
    tail = substr(expr, begin, len);

    if (debug == 1) printf("INSTR:head = <%s>\n", head);
    if (debug == 1) printf("INSTR:replace = <%s>\n", replace);
    if (debug == 1) printf("INSTR:tail = <%s>\n", tail);

    slot = slot_alloc("temp");
    printf("INSTR:ignore = negate(%s,%s)\n", slot, substr(replace,3,replace_len-2));

    reference = "@" slot;
    new_expr = head reference tail;
    print "EXPR:" new_expr;
    print "DONE:";
    fflush();
    next;
}

# DIFFERENCE:

$1 ~ /@[0-9]+-@[0-9]+/ {
    if (debug) printf("INSTR:difference\n");

    expr = $1;
    expr_len = length(expr);

    replace_pos = match(expr, /@[0-9]+-@[0-9]+/) - 1;
    replace_len = RLENGTH;

    begin = 1;
    len = replace_pos;
    head = substr(expr, begin, len);

    begin += replace_pos;
    len = replace_len;
    replace = substr(expr, begin, len);

    begin += replace_len;
    len = expr_len - begin + 1;
    tail = substr(expr, begin, len);

    if (debug) printf("INSTR:head = <%s>\n", head);
    if (debug) printf("INSTR:replace = <%s>\n", replace);
    if (debug) printf("INSTR:tail = <%s>\n", tail);

    split(replace, part, /-/);
    sub(/^@/, "", part[1]);
    sub(/^@/, "", part[2]);

    slot = slot_alloc("temp");
    printf("INSTR:ignore = difference(%s,%s,%s)\n", slot, part[1], part[2]);

    reference = "@" slot;
    new_expr = head reference tail;
    print "EXPR:" new_expr;
    print "DONE:";
    fflush();
    next;
}

$1 ~ /^[a-z]\/*[a-z]$/ {
    printf("INSTR:ignore = divide()\n");
    printf("EXPR:_%s_\n", 1);
    print "DONE:";
    fflush();
    next;
}

$1 ~ /^[a-z]\^*[a-z]$/ {
    printf("INSTR:ignore = conjugate()\n");
    printf("EXPR:_%s_\n", 1);
    print "DONE:";
    fflush();
    next;
}

' |&

{

echo '

define prompt() {
print "integer) "
}

define display(i) {
print a[i],"\n"
}

define initialize(i,n) {
a[i] = n;
}

define copy(i,j) {
a[i] = a[j];
}

define negate(i,j) {
a[i] = - a[j];
}

define sum(n,i,j) {
a[n] = a[i] + a[j];
}

define difference(n,i,j) {
a[n] = a[i] - a[j];
}

define product(n,i,j) {
a[n] = a[i] * a[j];
}

define divide() {
print "divide: "
}

define conjugate() {
print "conjugate: "
}

ignore = prompt()

'

while read
do
    expr=$REPLY

    passes=0

    while true
    do
	print -p -- ${expr}
	[[ $debug == 1 ]] && print "SENT <${expr}> to awk" > /dev/tty

	unset expr

	while read -pr result
	do
	    [[ $debug == 1 ]] && print "READ <${result}> from awk" > /dev/tty
	    [[ $result == @(EXPR:*)  ]] && expr=${result}
	    [[ $result == @(INSTR:*) && $debug == 0 ]] && print -r "${result#INSTR:}"
	    [[ $result == @(DONE:*)  ]] && break
	done

	[[ -z $expr ]] && break

	expr=${expr#EXPR:}
        passes=$((passes+1))
        [[ $passes -ge 20 ]] && break
    done

done

} |

if [[ $debug == 1 ]]
then
    cat
else
    bc -q
fi
